/*
  urlsnarf.c
  
  sniff the network for HTTP request URLs, output in CLF format.
  this is for demonstration purposes and educational use only.

  TODO: HTTP basic auth decode, non-GET HTTP methods
  
  Copyright (c) 1999 Dug Song <dugsong@monkey.org>
  All rights reserved, all wrongs reversed.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. The name of author may not be used to endorse or promote products
     derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
  THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  $Id: urlsnarf.c,v 1.7 2000/01/04 18:00:56 dugsong Exp $
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <nids.h>

#include "version.h"

#define int_ntoa(x)	inet_ntoa(*((struct in_addr *)&(x)))

void
usage(void)
{
  fprintf(stderr, "Usage: urlsnarf [-i interface]\n");
  exit(1);
}

/* XXX - blech. */
char *
get_request_time(void)
{
  static char tstr[32], sign;
  struct tm *t, gmt;
  time_t tt = time(NULL);
  int days, hours, tz, len;

  gmt = *gmtime(&tt);
  t = localtime(&tt);
  
  days = t->tm_yday - gmt.tm_yday;
  hours = ((days < -1 ? 24 : 1 < days ? -24 : days * 24) +
	   t->tm_hour - gmt.tm_hour);
  tz = hours * 60 + t->tm_min - gmt.tm_min;

  len = strftime(tstr, sizeof(tstr), "%e/%b/%Y:%X", t);
  if (len < 0 || len > sizeof(tstr) - 5)
    return (NULL);

  if (tz < 0) {
    sign = '-';
    tz = -tz;
  }
  else sign = '+';
  
  snprintf(tstr + len, sizeof(tstr) - len, " %c%.2d%.2d",
	   sign, tz / 60, tz % 60);
  
  return (tstr);
}

void
sniff_http_client(struct tcp_stream *ts, void **yoda)
{
  char *p, *buf;
  char *requrl, *vhost, *reqtime;
  int len;

  /* Only handle HTTP client traffic. */
  if (ts->addr.dest != 80)
    return;

  switch (ts->nids_state) {
  case NIDS_JUST_EST:
    /* Collect data. */
    ts->server.collect = 1;
    
  case NIDS_DATA:
    buf = ts->server.data;
    if ((len = ts->server.count - ts->server.offset) <= 0)
      return;
    buf[len] = '\0';
    
    /* Save until we get an entire request. */
    p = buf;
    if ((p = strstr(buf, "\r\n\r\n")) == NULL) {
      nids_discard(ts, 0);
      return;
    }
    /* Complete request seen. */
    nids_discard(ts, (p + 4) - buf);
    *p = '\0';

    requrl = vhost = reqtime = NULL;
    
    for (p = strtok(buf, "\r\n"); p != NULL; p = strtok(NULL, "\r\n")) {
      if (strncasecmp(p, "GET ", 4) == 0) {
	requrl = p + 4;
	if (strncmp(requrl, "http://", 7) == 0)
	  vhost = "";
      }
      else if (strncasecmp(p, "Host: ", 6) == 0)
	vhost = p + 6;
    }
    reqtime = get_request_time();
    
    if (requrl != NULL && vhost != NULL && reqtime != NULL) {
      printf("%s - - [%s] \"GET %s%s%s\" - -\n", int_ntoa(ts->addr.saddr),
	     reqtime, (*vhost == '\0' ? vhost : "http://"), vhost, requrl);
    }
    break;
    
  default:
    break;
  }
}

int
main(int argc, char *argv[])
{
  char c;
  
  while ((c = getopt(argc, argv, "i:h?V")) != EOF) {
    switch (c) {
    case 'i':
      nids_params.device = optarg;
      break;
    case 'V':
      fprintf(stderr, "Version: %s\n", VERSION);
      usage();
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0)
    usage();
  
  if (!nids_init()) {
    fprintf (stderr, "%s\n", nids_errbuf);
    exit(1);
  }
  nids_register_tcp(sniff_http_client);

  nids_run();

  /* NOTREACHED */
  exit(0);
}

/* 5000. */
